import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np 
from scipy.stats import uniform
import six
import collections
import copy
from scipy.stats import norm
from itertools import product
from sklearn.metrics import mean_squared_error
from random import sample 
from sklearn.datasets import make_moons
from collections import defaultdict
SKIP_TYPES = six.string_types



class FastTensorDataLoader:
    """
    A DataLoader-like object for a set of tensors that can be much faster than
    TensorDataset + DataLoader because dataloader grabs individual indices of
    the dataset and calls cat (slow).
    Source: https://discuss.pytorch.org/t/dataloader-much-slower-than-manual-batching/27014/6
    """
    def __init__(self, *tensors, tensor_names, batch_size=32, shuffle=False):
        """
        Initialize a FastTensorDataLoader.
        :param *tensors: tensors to store. Must have the same length @ dim 0.
        :param tensor_names: name of tensors (for feed_dict)
        :param batch_size: batch size to load.
        :param shuffle: if True, shuffle the data *in-place* whenever an
            iterator is created out of this object.
        :returns: A FastTensorDataLoader.
        """
        assert all(t.shape[0] == tensors[0].shape[0] for t in tensors)
        self.tensors = tensors
        self.tensor_names = tensor_names

        self.dataset_len = self.tensors[0].shape[0]
        self.batch_size = batch_size
        self.shuffle = shuffle

        # Calculate # batches
        n_batches, remainder = divmod(self.dataset_len, self.batch_size)
        if remainder > 0:
            n_batches += 1
        self.n_batches = n_batches

    def __iter__(self):
        if self.shuffle:
            r = torch.randperm(self.dataset_len)
            self.tensors = [t[r] for t in self.tensors]
        self.i = 0
        return self

    def __next__(self):
        if self.i >= self.dataset_len:
            raise StopIteration
        batch = {}
        for k in range(len(self.tensor_names)):
            batch.update({self.tensor_names[k]: self.tensors[k][self.i:self.i+self.batch_size]})
        self.i += self.batch_size
        return batch
        

    def __len__(self):
        return self.n_batches

class Meters:
    def __init__(self):
        self.data = defaultdict(list)

    def update(self, updates=None, value=None, n=1, **kwargs):
        """
        Example:
            >>> meters.update(key, value)
            >>> meters.update({key1: value1, key2: value2})
            >>> meters.update(key1=value1, key2=value2)
        """
        if updates is None:
            updates = {}
        if updates is not None and value is not None:
            updates = {updates: value}
        updates.update(kwargs)
        for k, v in updates.items():
            self.data[k].append(v)

    @property
    def avg(self):
        averages = {}
        for key, values in self.data.items():
            averages[key] = sum(values) / len(values)
        return averages

    def get_values(self, key):
        return self.data[key]

def get_activation(act):
    if isinstance(act, nn.Module):
        return act
    assert type(act) is str, 'Unknown type of activation: {}.'.format(act)
    act_lower = act.lower()
    if act_lower == 'relu':
        return nn.ReLU(True)
    elif act_lower == 'selu':
        return nn.SELU(True)
    elif act_lower == 'sigmoid':
        return nn.Sigmoid()
    elif act_lower == 'tanh':
        return nn.Tanh()
    else:
        try:
            return getattr(nn, act)
        except AttributeError:
            raise ValueError('Unknown activation function: {}.'.format(act))


def get_optimizer(optimizer, model, *args, **kwargs):
    if isinstance(optimizer, (optim.Optimizer)):
        return optimizer

    if type(optimizer) is str:
        try:
            optimizer = getattr(optim, optimizer)
        except AttributeError:
            raise ValueError('Unknown optimizer type: {}.'.format(optimizer))
    return optimizer(filter(lambda p: p.requires_grad, model.parameters()), *args, **kwargs)
    

def stmap(func, iterable):
    if isinstance(iterable, six.string_types):
        return func(iterable)
    elif isinstance(iterable, (collections.abc.Sequence, collections.UserList)):
        return [stmap(func, v) for v in iterable]
    elif isinstance(iterable, collections.abc.Set):
        return {stmap(func, v) for v in iterable}
    elif isinstance(iterable, (collections.abc.Mapping, collections.UserDict)):
        return {k: stmap(func, v) for k, v in iterable.items()}
    else:
        return func(iterable)




def _as_numpy(o):
    from torch.autograd import Variable
    if isinstance(o, SKIP_TYPES):
        return o
    if isinstance(o, Variable):
        o = o
    if torch.is_tensor(o):
        return o.cpu().numpy()
    return np.array(o)


def as_numpy(obj):
    return stmap(_as_numpy, obj)


def _as_float(o):
    if isinstance(o, SKIP_TYPES):
        return o
    if torch.is_tensor(o):
        return o.item()
    arr = as_numpy(o)
    assert arr.size == 1
    return float(arr)


def as_float(obj):
    return stmap(_as_float, obj)




# Create a simple dataset
def create_twomoon_dataset(n, p):
    relevant, y = make_moons(n_samples=n, shuffle=True, noise=0.1, random_state=None)
    noise_vector = norm.rvs(loc=0, scale=1, size=[n,p-2])
    X = np.concatenate([relevant, noise_vector], axis=1)
    return dict(X=X, y=y.flatten())



def get_model_size(model_path):
    size = []
    for j, _model in enumerate(model_path):
        if hasattr(_model.mlp[0], "__getitem__"):
            weight = _model.mlp[0][0].weight
        else:
            weight = _model.mlp[0].weight
        if torch.isnan(weight).any():
            size.append(np.nan)
        else:
            var_sel= torch.sum(torch.abs(weight), 0)>1e-6
            size.append(int(sum(var_sel)))
    return np.array(size)


def get_param_combination(param_grid):
    param_list = [param_grid[key] for key in param_grid]
    comb_list = []
    for comb in product(*param_list):
        comb_dict = {key: value for key, value in zip(param_grid.keys(), comb)}
        comb_list.append(comb_dict)
    return comb_list
         



def get_mse(model, X, y, best=True):
    if best:
        model._model = model.best_model
    if model.drop_input:
        X = X[:,model._index_path[model.best_lam_ind]]
    y_hat= model.predict(X)
    return mean_squared_error(y, y_hat)


def nanargmax_safe(arr, axis=None):
    if np.isnan(arr).all():
        return len(arr.flatten())-1
    else:
        return np.nanargmax(arr, axis=axis)
    

def _standard_truncnorm_sample(lower_bound, upper_bound, sample_shape=torch.Size()):
    r"""
    Implements accept-reject algorithm for doubly truncated standard normal distribution.
    (Section 2.2. Two-sided truncated normal distribution in [1])
    [1] Robert, Christian P. "Simulation of truncated normal variables." Statistics and computing 5.2 (1995): 121-125.
    Available online: https://arxiv.org/abs/0907.4010
    Args:
        lower_bound (Tensor): lower bound for standard normal distribution. Best to keep it greater than -4.0 for
        stable results
        upper_bound (Tensor): upper bound for standard normal distribution. Best to keep it smaller than 4.0 for
        stable results
    """
    x = torch.randn(sample_shape)
    done = torch.zeros(sample_shape).byte() 
    while not done.all():
        proposed_x = lower_bound + torch.rand(sample_shape) * (upper_bound - lower_bound)
        if (upper_bound * lower_bound).lt(0.0):  # of opposite sign
            log_prob_accept = -0.5 * proposed_x**2
        elif upper_bound < 0.0:  # both negative
            log_prob_accept = 0.5 * (upper_bound**2 - proposed_x**2)
        else:  # both positive
            assert(lower_bound.gt(0.0))
            log_prob_accept = 0.5 * (lower_bound**2 - proposed_x**2)
        prob_accept = torch.exp(log_prob_accept).clamp_(0.0, 1.0)
        accept = torch.bernoulli(prob_accept).byte() & ~done
        if accept.any():
            accept = accept.bool()
            x[accept] = proposed_x[accept]
            accept = accept.byte()
            done |= accept
    return x

def simulate_data(n_samples, n_features, scale=2, shape=2, data_type="cox", censoring_rate=0.1,
 times_distribution="weibull"):
    """  n_samples=10
        n_features=5
        scale=0.5
        shape=0.5
        censoring_factor=2
        times_distribution="weibull" """  
    mu=np.zeros(n_features)
    cov=np.diag(np.ones(n_features))
    X=np.random.multivariate_normal(mean=mu, cov=cov, size=n_samples)
    u= X[:,0] + X[:,1]* X[:,0]+ np.log(np.abs(X[:,1])+0.1)+ np.exp(X[:,2]+X[:,3])  
         
    if data_type == "cox":
        # Simulation of true times
        E = np.random.exponential(scale=1., size=n_samples)
        E *= np.exp(-u)
        if times_distribution == "weibull":
            T = (E / scale ) ** (1. / shape)
        else:
            # There is not point in this test, but let's do it like that
            # since we're likely to implement other distributions
            T = (E / scale ) ** (1. / shape)
        censor_index = sample(range(n_samples), int(n_samples * censoring_rate))
        t=copy.deepcopy(T) 
        t[censor_index] = np.random.uniform(0, T[censor_index])
        e=(T==t)
        return dict(X=X, T=t, E=e)
    elif data_type == "regression":
        eps = np.random.normal(loc=0, scale=1, size=n_samples)
        y = u + eps
        return dict(X=X,y=y)
    elif data_type == "classification":
        p = 1/(1+np.exp(-u))
        y = np.random.binomial(1,p=p)
        return dict(X=X, y=y.flatten())
    

def simulate_bivariate_cox_data(n_samples, n_features, scale=2, shape=2, censoring_rate=0.1,
 times_distribution="weibull", case="LN", theta=2):
    """  n_samples=10
        n_features=5
        scale=0.5
        shape=0.5
        censoring_factor=2
        times_distribution="weibull" """  
    pnum = 5
    mu=np.zeros(n_features)
    cov=np.diag(np.ones(n_features))
    X=np.random.multivariate_normal(mean=mu, cov=cov, size=n_samples)
    if case=="NL1":
        u= 3*X[:,0]+5*X[:,1]**2 +4*X[:,0]*X[:,1]+ X[:,2]**2 + 3*X[:,3] + 5* X[:,3]*X[:,4]
        u2=3*X[:,2]+2*X[:,1] +4*X[:,2]*X[:,1]+ X[:,0]**2 + 5*X[:,4] + 3* X[:,4]*X[:,3]
    elif case == "NL2":
        u= X[:,0] + X[:,1]* X[:,0]+ np.log(np.abs(X[:,1])+0.1)+np.exp(X[:,2]+X[:,3]) + X[:,4]**2 
        u2=X[:,0]**2 + 4*X[:,1]+ X[:,1]*X[:,2] + 3*np.log(np.abs(X[:,2])+0.5) + 2*np.exp(X[:,3]+X[:,4])
    elif case == "LN":
        beta = np.array([0.729,  0.493,   0.200, -1.625,  -1.081])
        beta2 = np.array([0.701, -1.095,  0.435,  0.080,  0.053])
        u = np.dot(X[:,range(pnum)], beta)
        u2 = np.dot(X[:,range(pnum)], beta2)

    # Generate random uniform samples U1 and U2
    U1 = uniform.rvs(size=n_samples)
    V2 = uniform.rvs(size=n_samples)

    # Apply the inverse Clayton copula CDF to get copula-based samples
    U2=(U1**(-theta)*(V2**(-theta/(1+theta))-1)+1)**(-1/theta)

    # Simulation of true times for first outcome
    E=-np.log(1 - U1)
    #E = np.random.exponential(scale=1., size=n_samples)
    E *= np.exp(-u)
    if times_distribution == "weibull":
        T = (E / scale ) ** (1. / shape)
    else:
        # There is not point in this test, but let's do it like that
        # since we're likely to implement other distributions
        T = (E / scale ) ** (1. / shape)
    censor_index = sample(range(n_samples), int(n_samples * censoring_rate))
    t=copy.deepcopy(T) 
    t[censor_index] = np.random.uniform(0, T[censor_index])
    e=(T==t)

    # Simulation of true times for second outcome
    E2=-np.log(1 - U2)
    #E2 = np.random.exponential(scale=1., size=n_samples)
    E2 *= np.exp(-u2)
    if times_distribution == "weibull":
        T2 = (E2 / scale ) ** (1. / shape)
    else:
        # There is not point in this test, but let's do it like that
        # since we're likely to implement other distributions
        T2 = (E2 / scale ) ** (1. / shape)
    censor_index = sample(range(n_samples), int(n_samples * censoring_rate))
    t2=copy.deepcopy(T2) 
    t2[censor_index] = np.random.uniform(0, T2[censor_index])
    e2=(T2==t2)
    return dict(X=X, T=t, E=e, T2=t2, E2=e2)



import torch.nn as nn
import torch
from gcrnet.utils import get_activation

__all__ = [
    'MLPLayer', 'MLPModel'
]

class MLPLayer(nn.Sequential):
    def __init__(self, in_features, out_features, bias=True, activation=None):
        modules = [nn.Linear(in_features, out_features, bias=bias)]
        if activation is not None and activation is not False:
            modules.append(get_activation(activation))
        super().__init__(*modules)

class ParallelMLPLayer(nn.Module):
    def __init__(self, in_features, out_features, n_para, bias=True, activation=None):
        super(ParallelMLPLayer, self).__init__()
        modules = []
        self.n_para = n_para
        self.in_features = in_features
        for _ in range(n_para):
            layer = MLPLayer(in_features, out_features, activation=activation, bias=bias) 
            modules.append(layer)
        self.layers = nn.ModuleList(modules)

    def forward(self, x):
        outputs = []
        for i, layer in enumerate(self.layers):
            start_idx = i * self.in_features
            end_idx = (i + 1) * self.in_features
            outputs.append(layer(x[:, start_idx:end_idx]))
        return torch.cat(outputs, dim=1)
    

class MLPModel(nn.Module):
    def __init__(self, input_dim, output_dim, hidden_dims,  activation='relu', flatten=True, output_bias=False):
        super().__init__()

        if hidden_dims is None:
            hidden_dims = []
        elif type(hidden_dims) is int:
            hidden_dims = [hidden_dims]
        nr_hiddens = len(hidden_dims)


        n_para=None
        if output_dim>1 and nr_hiddens>0:
            n_para=output_dim
            output_dim=1

        dims = [input_dim]
        dims.extend(hidden_dims)
        dims.append(output_dim)
        modules = []

        
        if n_para is None: 
            for i in range(nr_hiddens):
                layer = MLPLayer(dims[i], dims[i+1],  activation=activation)
                modules.append(layer)
            layer = nn.Linear(dims[-2], dims[-1], bias=output_bias)
        else:
            for i in range(nr_hiddens):
                if i==0:
                    layer = MLPLayer(dims[i], dims[i+1]*n_para,  activation=activation)
                else:
                    layer = ParallelMLPLayer(dims[i], dims[i+1], n_para=n_para, activation=activation)
                modules.append(layer)
            layer =  ParallelMLPLayer(dims[-2], dims[-1], n_para=n_para, bias=output_bias)
        modules.append(layer)
        self.mlp = nn.Sequential(*modules)
        self.flatten = flatten

    def forward(self, input):
        if self.flatten:
            input = input.view(input.size(0), -1)
        return self.mlp(input)

    
    def get_input_layer(self):
        if hasattr(self.mlp[0], "__getitem__"):
            return self.mlp[0][0]
        else:
            return self.mlp[0]
        
    def set_input_layer(self, new_input_layer):
        if hasattr(self.mlp[0], "__getitem__"):
            self.mlp[0][0] = new_input_layer
        else:
            self.mlp[0] = new_input_layer


import torch.nn as nn
import torch
import numpy as np

from gcrnet.mlp import MLPModel
from gcrnet.losses import PartialLogLikelihood
from gcrnet.penalty import *




class ConcaveRegularMLPModel(MLPModel):
    def __init__(self, input_dim, output_dim, hidden_dims,  activation='relu',
                 outer_penalty=None, inner_penalty=None, lam=0, alpha=0, gamma=1, output_bias=False):
        super().__init__(input_dim, output_dim, hidden_dims, activation=activation, output_bias=output_bias)
        self.outer_penalty = outer_penalty
        self.inner_penalty = inner_penalty
        self.lam=lam
        self.alpha=alpha
        self.gamma=gamma
        if outer_penalty is not None and inner_penalty is not None:
            assert type(outer_penalty) is str, 'Unknown type of outer penalty: {}.'.format(outer_penalty)
            assert type(inner_penalty) is str, 'Unknown type of inner penalty: {}.'.format(inner_penalty)

    def concave_soft_thresh_update(self):
        if self.outer_penalty is None or self.inner_penalty is None:
            return 0 
        len_sparse_layer=1
        penalty_val = 0
        with torch.no_grad():
            for i in range(len_sparse_layer):
                if hasattr(self.mlp[i], "__getitem__"):
                    for j in range(self.mlp[i][0].weight.shape[1]):
                        self.mlp[i][0].weight.data[:,j]=ConcaveSoftThresh(self.mlp[i][0].weight[:,j], lam=self.lam, gamma=self.gamma, 
                                                            outer_penalty=self.outer_penalty, inner_penalty=self.inner_penalty)
                        penalty_val = penalty_val +get_penalty_val(self.mlp[i][0].weight.data[:,j], lam=self.lam, 
                                                            outer_penalty=self.outer_penalty, inner_penalty=self.inner_penalty)
                        for a in self.mlp[i][0].weight.data[:,j]:
                            if torch.isnan(a).any():
                                raise ValueError("NaNs detected in inputs, please correct or drop.")
                else:
                    for j in range(self.mlp[i].weight.shape[1]):
                        self.mlp[i].weight.data[:,j]=ConcaveSoftThresh(self.mlp[i].weight[:,j], lam=self.lam, gamma=self.gamma,
                                                            outer_penalty=self.outer_penalty, inner_penalty=self.inner_penalty)
                        val = get_penalty_val(self.mlp[i].weight.data[:,j], lam=self.lam, 
                                                            outer_penalty=self.outer_penalty, inner_penalty=self.inner_penalty)                 
                        penalty_val = penalty_val + val
        return self.gamma*penalty_val 
                
    def get_concave_penalty_val(self):
        if self.outer_penalty is None or self.inner_penalty is None:
            return 0 
        len_sparse_layer=1
        penalty_val = 0
        with torch.no_grad():
            for i in range(len_sparse_layer):
                if hasattr(self.mlp[i], "__getitem__"):
                    for j in range(self.mlp[i][0].weight.shape[1]):
                        penalty_val = penalty_val +get_penalty_val(self.mlp[i][0].weight.data[:,j], lam=self.lam, 
                                                            outer_penalty=self.outer_penalty, inner_penalty=self.inner_penalty)
                else:
                    for j in range(self.mlp[i].weight.shape[1]):
                        val = get_penalty_val(self.mlp[i].weight.data[:,j], lam=self.lam, 
                                                            outer_penalty=self.outer_penalty, inner_penalty=self.inner_penalty)                 
                        penalty_val = penalty_val + val
        return self.gamma* penalty_val 
        

    def l2_regularize(self):
        if self.alpha != 0:
            return self.alpha * sum(p.pow(2.0).sum() for p in self.parameters())
        return 0

### Regression Model ###
class ConcaveRegularMLPRegressionModel(ConcaveRegularMLPModel):
    def __init__(self, input_dim, output_dim, hidden_dims,  activation='relu',
                    outer_penalty=None, inner_penalty=None, lam=0, alpha=0, gamma=1):
        super().__init__(input_dim, output_dim, hidden_dims, activation=activation,
         outer_penalty=outer_penalty, inner_penalty=inner_penalty, lam=lam, alpha=alpha, gamma=gamma, output_bias=True)
        self.loss = nn.MSELoss()

    def forward(self, feed_dict):
        pred = super().forward(feed_dict['X'])
        if pred.shape[1]==1:
            pred = pred.view(pred.size(0))
        if self.training:
            loss = self.loss(pred, feed_dict['y']) + self.l2_regularize()
            return loss, dict(), dict()
        else:
            return dict(pred=pred)

### Classification Model ###
class ConcaveRegularMLPClassificationModel(ConcaveRegularMLPModel):
    def __init__(self, input_dim, output_dim, hidden_dims, activation='relu',
                    outer_penalty=None, inner_penalty=None, lam=0, alpha=0, gamma=1):
        super().__init__(input_dim, output_dim, hidden_dims, activation=activation,
         outer_penalty=outer_penalty, inner_penalty=inner_penalty, lam=lam, alpha=alpha,  gamma=gamma, output_bias=True)
        # self.softmax = nn.Softmax(dim=1)
        # self.loss = nn.CrossEntropyLoss()
        self.sigmoid = nn.Sigmoid()
        self.loss = nn.BCELoss()

    def forward(self, feed_dict):
        logits = super().forward(feed_dict['X'])
        if logits.shape[1]==1:
            logits = logits.view(logits.size(0))
        if self.training:
            loss = self.loss(self.sigmoid(logits), feed_dict['y'].to(torch.float32) ) + self.l2_regularize()
            return loss, dict(), dict()
        else:
            return self._compose_output(logits)

    def _compose_output(self, logits):
        value = self.sigmoid(logits)
        pred = np.array(value > 0.5)*1
        return dict(prob=value, pred=pred, logits=logits)
### Cox Model
class ConcaveRegularMLPCoxModel(ConcaveRegularMLPModel):
    def __init__(self, input_dim, output_dim, hidden_dims,  activation='relu',
                    outer_penalty=None, inner_penalty=None, lam=0, alpha=0, gamma=1):
        super().__init__(input_dim, output_dim, hidden_dims, activation=activation,
         outer_penalty=outer_penalty, inner_penalty=inner_penalty, lam=lam, alpha=alpha, gamma=gamma,  output_bias=False)
        self.loss = PartialLogLikelihood 
        self.noties = 'noties'

    def forward(self, feed_dict):
        logits = super().forward(feed_dict['X'])
        if self.training:
            loss = self.loss(logits, feed_dict['E'].reshape(-1, 1), self.noties) + self.l2_regularize()
            return loss, logits, dict()
        else:
            return dict(logits=logits)

## Bivariate Cox Model
class ConcaveRegularMLPBivariateCoxModel(ConcaveRegularMLPModel):
    def __init__(self, input_dim, hidden_dims,  activation='relu',
                    outer_penalty=None, inner_penalty=None, lam=0, alpha=0, gamma=1):
        super().__init__(input_dim=input_dim, output_dim=2, hidden_dims=hidden_dims, activation=activation,
         outer_penalty=outer_penalty, inner_penalty=inner_penalty, lam=lam, alpha=alpha, gamma=gamma, output_bias=False)
        self.loss = PartialLogLikelihood 
        self.noties = 'noties'
        self.sort_id2 = None

    def forward(self, feed_dict):
        logits = super().forward(feed_dict['X'])
        if self.training:
            if self.sort_id2 is None:
                self.sort_id2 = torch.flip(torch.argsort(feed_dict['T2']), dims=[0])
            ordered_E2 = feed_dict['E2'][self.sort_id2]
            loss = self.loss(logits[:,0], feed_dict['E'].reshape(-1, 1), self.noties) + self.loss(logits[self.sort_id2,1], ordered_E2.reshape(-1, 1), self.noties) + self.l2_regularize()
            #loss = self.loss(logits[self.sort_id2,1], ordered_E2.reshape(-1, 1), self.noties) + self.l2_regularize()
            #loss = self.loss(logits[:,0], feed_dict['E'].reshape(-1, 1), self.noties) + self.l2_regularize()
            return loss, logits, dict()
        else:
            return dict(logits=logits)




    


import torch 
import numpy as np
from lifelines.utils import concordance_index

def PartialLogLikelihood(logits, fail_indicator, ties):
    '''
    fail_indicator: 1 if the sample fails, 0 if the sample is censored.
    logits: raw output from model 
    ties: 'noties' or 'efron' or 'breslow'
    '''
    logL = 0
    if ties == 'noties':
        likelihood = logits - torch.logcumsumexp(logits, dim=0)
        # dimension for E: np.array -> [None, 1]
        uncensored_likelihood = likelihood.squeeze() * fail_indicator.squeeze()
        logL = -torch.sum(uncensored_likelihood)
    else:
        raise NotImplementedError()
    # negative average log-likelihood
    observations = torch.sum(fail_indicator, 0)
    return 1.0*logL / observations


def calc_concordance_index(logits, fail_indicator, fail_time):
    """
    Compute the concordance-index value.
    Parameters:
        label_true: dict, like {'e': event, 't': time}, Observation and Time in survival analyze.
        y_pred: np.array, predictive proportional risk of network.
    Returns:
        concordance index.
    """
    hr_pred = -logits 
    try:
        for a in [fail_time, hr_pred, fail_indicator]:
            if np.isnan(a).any():
                raise ValueError("NaNs detected in inputs, please correct or drop.")
    except ValueError as e:
        print(e)
        return 0.5
    ci = concordance_index(fail_time,
                            hr_pred,
                            fail_indicator)
    return ci
